/*
 * Java Othello Openings Applet
 * Copyright (C) 2000 Bill Lorton
 * email: hogarth1@yahoo.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

import java.awt.*;
import java.awt.event.*;

/**
 *
 * Creates and handles the continuations only checkbox, the list and
 * the predecessor button.  A linked list of the items that are currently
 * in the list is kept in this class.
 *
 * @version 1.00
 * @author Bill Lorton
 *
 */
class ListPanel extends Panel {
  private Othello _applet;

  private GridBagLayout _gridbag = new GridBagLayout();
  private List          _list;
  private Checkbox      _cont_cb;
  private Button        _pred_btn;

  private ListItem _items,_last_item;

  /**
   *
   * Recursively adds all of the continations under the given node.
   *
   * @param node where to begin the search for openings
   *
   */
  protected void _addContinuationItems(OTNode node)
  {
    if (node==null)
      return;

    if (node.data()!=null) {
      addItem(node);
      _addContinuationItems((OTNode)node.next());
      return;
    }

    _addContinuationItems((OTNode)node.child());
    _addContinuationItems((OTNode)node.next());
  }

  // PUBLIC

  /**
   *
   * Listener for the continuations only check box.
   *
   * Note that javadoc 1.1.x does not generate docs for inner classes.
   *
   * @version 1.00
   * @author Bill Lorton
   *
   */
  public class HandleCheckbox implements ItemListener {
    public HandleCheckbox() { }

    public void itemStateChanged(ItemEvent e)
    {
      boolean set = e.getStateChange()==e.SELECTED;

      _applet.continuationsOnlyChangedCallback(set);
      _pred_btn.setEnabled(set && !_applet.initialSetup());
    }
  }

  /**
   *
   * Listener for the list.
   *
   * Note that javadoc 1.1.x does not generate docs for inner classes.
   *
   * @version 1.00
   * @author Bill Lorton
   *
   */
  public class HandleList implements ItemListener {
    public HandleList() { }

    public void itemStateChanged(ItemEvent e)
    {
      if (e.getStateChange()==e.DESELECTED)
	return; // don't care about deselection of items

      ListItem item = _items;
      int      n    = _list.getSelectedIndex();

      // find the associated item in the linked list

      while (n>0) {
	item = (ListItem)item.next();
	--n;
      }

      boolean cont_set = _cont_cb.getState();

      _applet.listSelectCallback(item.otNode(),cont_set);

      // an item has just been selected. if the continuation checkbox is
      // set then there must be a predecessor.  update the enabled state
      // of the button.

      if (cont_set)
	_pred_btn.setEnabled(true);
    }
  }

  /**
   *
   * Listener for the predecessor button.
   *
   * Note that javadoc 1.1.x does not generate docs for inner classes.
   *
   * @version 1.00
   * @author Bill Lorton
   *
   */
  public class HandleButton implements ActionListener {
    HandleButton() { }

    public void actionPerformed(ActionEvent e)
    {
      if (!_applet.previousOpening())
    	_pred_btn.setEnabled(false);
    }
  }

  /**
   *
   * Creates the interface components.
   *
   */
  public ListPanel(Othello app)
  {
    _applet    = app;
    _items     = null;
    _last_item = null;
    GBHelper gbh = new GBHelper();

    setLayout(_gridbag);

    _cont_cb = new Checkbox();
    _cont_cb.addItemListener(new HandleCheckbox());
    _cont_cb.setLabel("Continuations Only");
    _cont_cb.setState(false);
    gbh.constrainFill(this,_cont_cb,0,0,1,1);

    _list = new List(10,false);
    _list.addItemListener(new HandleList());
    _list.setMultipleMode(false);
    gbh.constrainFill(this,_list,0,1,1,2);

    _pred_btn = new Button("Predecessor");
    _pred_btn.addActionListener(new HandleButton());
    _pred_btn.setEnabled(false);
    gbh.constrainFill(this,_pred_btn,0,4,1,1,5,0,0,0);

    validate(); // force the layout of the components
  }


  /**
   *
   * Adds the given node's opening to the list.  The node must contain
   * an opening.
   *
   * @param node the node to add to the list
   *
   */
  public void addItem(OTNode node)
  {
    ListItem ln = new ListItem();

    ln.setOTNode(node);

    if (_items==null)
      _items = ln;
    else
      _last_item.setNext(ln);

    _last_item = ln;

    _list.addItem(node.data().name());
  }


  /**
   *
   * Makes the given node's opening selected in the list.
   *
   * @param node the node to select
   *
   */
  public void selectItem(OTNode node)
  {
    ListItem ln   = _items;
    int      indx = 0;

    while (ln!=null) {
      if (ln.otNode().data()==node.data())
        break;

      ++indx;
      ln = (ListItem)ln.next();
    }

    if (ln!=null) {  // not in the list..  this should never happen
      _list.select(indx);
      _list.makeVisible(indx);
    }
  }


  /**
   *
   * Adds the continuations of the given node's opening to the list.  The node 
   * must contain an opening.
   *
   * @param node the node that will have its continuations added
   *
   */
  public void addContinuationItems(OTNode node)
  {
    _addContinuationItems((OTNode)node.child());
  }


  /**
   *
   * Removes all items from the list.
   *
   */
  public void removeAll()
  {
    _list.removeAll();
    _items     = null;
    _last_item = null;
  }
}
