/*
 * Java Othello Openings Applet
 * Copyright (C) 2000 Bill Lorton
 * email: hogarth1@yahoo.com
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

import java.awt.*;
import java.awt.event.*;
import java.applet.Applet;
import java.io.*;
import java.net.*;
import java.util.StringTokenizer;

/**
 *
 * Java Othello Openings Applet
 *
 * @version 1.00
 * @author Bill Lorton
 *
 */

public class Othello extends Applet {
  static private final int A = 0;
  static private final int B = 1;
  static private final int C = 2;
  static private final int D = 3;
  static private final int E = 4;
  static private final int F = 5;
  static private final int G = 6;
  static private final int H = 7;

  /**
   *
   * The top node of the frequency sorted opening nodes.  This node does not
   * itself contain data.
   *
   */
  protected OTNode _alpha_openings;

  /**
   *
   * The root node of a hierarchical tree of nodes that is organized by move.
   * This node does not contain any move data.
   *
   */
  protected OTNode _openings;

  /**
   *
   * The current opening.
   *
   */
  protected OTNode _opening;

  /**
   *
   * The base orientation lookup table.  This table is automatically
   * generated and has values 1-64.  The lookup table arrays are
   * arrays of 64 integers which represent the 8x8 board.  Row one of
   * the board is index position 0-7 of the array, row two is 8-15, etc.
   * The arrays are lookup tables that determine how discs are to be
   * displayed in the applet.  To determine what row/col a disc should
   * be displayed in on the board, the disc's position is converted from a
   * row/column value into a single value using _translateRowColumn(r,c).
   * This single value is used to index _curr_lookup which is the pointer
   * to the currently selected orientation array.  The value that is
   * returned is then converted back to a row and column value and the
   * disc is drawn.
   *
   */
  protected int _o1_lookup[];

  /**
   *
   * A lookup table for the second board orientation.  This table is
   * read from "orientations.txt".
   *
   */
  protected int _o2_lookup[];

  /**
   *
   * A lookup table for the third board orientation.  This table is
   * read from "orientations.txt".
   *
   */
  protected int _o3_lookup[];

  /**
   *
   * A lookup table for the fourth board orientation.  This table is
   * read from "orientations.txt".
   *
   */
  protected int _o4_lookup[];

  /**
   *
   * The current board orientation lookup table.  This points to the
   * lookup table that cooresponds to the currently selected board
   * orientation.
   *
   */
  protected int _curr_lookup[];

  /**
   *
   * Stores the moves of the selected opening that have been made.
   *
   */
  protected String _played_moves;

  /**
   *
   * Stores the moves of the selected opening that have not yet been made.
   *
   */
  protected String _unplayed_moves;

  private Button        _first_btn,_prev_btn,_next_btn,_last_btn,_play_btn;
  private Button        _stop_btn;
  private Label         _name_label,_moves_label,_frequency_label;
  private Panel         _right_panel;
  private CheckboxGroup _orientation_group;
  private Checkbox      _o1_box;
  private Checkbox      _o2_box;
  private Checkbox      _o3_box;
  private Checkbox      _o4_box;

  /**
   *
   * Handles the drawing of the board.
   *
   */
  protected Board _board;

  /**
   *
   * Handles the updating and manipulation of the list.
   *
   */
  protected ListPanel _list_panel;

  /**
   *
   * Thread object for playing the moves.
   *
   */
  protected PlayerThread _player_thread;


// PROTECTED METHODS

  /**
   *
   * Parses the opening name from the string.
   *
   * @param st the string to parse
   * @return the name of the opening
   *
   **/
  protected String _parseName(StringTokenizer st)
  {
    String  name         = new String("");
    String  last_token   = new String("");
    boolean name_started = false;

    while (st.hasMoreTokens()) {
      String token = st.nextToken();

      if (!name_started && !token.equals(" ") && !token.equals("\t"))
        name_started = true;

      if (name_started) {
        if (token.equals(" ") && last_token.equals(" "))
          break;

        if (token.equals("(t3)"))
          break;

        if (token.equals("\n")) // for sequence section
          break;

        if (!token.equals("*"))
          name += token;

        last_token = token;
      }
    }

    name = name.trim();
    return name;
  }


  /**
   *
   * Parses the orientation array from a string.
   *
   * @param st    the string tokenizer to parse
   * @param array the resulting array
   *
   */
  protected void _parseOrientation(StringTokenizer st,int array[])
  {
    for (int i=0;i<64;++i) {
      array[i] = Integer.parseInt(st.nextToken());
    }
  }


  /**
   *
   * Reads "orientations.txt" which contains 3 rows of 64 integer values
   * that form three of the orientation lookup tables (_o2_lookup,_o3_lookup,
   * and _o4_lookup).  The fourth lookup table (_o1_lookup) is automatically
   * generated as values 1-64.  The current lookup table is initialized to
   * _o1_lookup.
   *
   */
  protected void _readOrientations()
  {
    URL fileUrl;

    String data = "";

    try {
      showStatus("Retrieving orientations.txt...");
      fileUrl = new URL(getCodeBase(),"orientations.txt");
      InputStream infile = fileUrl.openStream();

      BufferedReader ds = new BufferedReader(new InputStreamReader(infile));

      for (;;) {
        String line = ds.readLine();

        if (line==null)
          break;

        data += line.trim() + "\n";
      }
    }
    catch (MalformedURLException mue) {
      System.out.println("Bad filename.");
      mue.printStackTrace();
    } catch (IOException ioe) {
      System.out.println("Data IO exception.");
      ioe.printStackTrace();
    }

    _o1_lookup = new int[64];
    _o2_lookup = new int[64];
    _o3_lookup = new int[64];
    _o4_lookup = new int[64];

    for (int i=0;i<64;++i)
      _o1_lookup[i] = i+1;

    StringTokenizer st  = new StringTokenizer(data);

    _parseOrientation(st,_o2_lookup);
    _parseOrientation(st,_o3_lookup);
    _parseOrientation(st,_o4_lookup);
    _curr_lookup = _o1_lookup;
  }


  /**
   *
   * Reads and processes "openings.txt".
   *
   */
  protected void _readOpenings()
  {
    URL fileUrl;
    String data = "";  // the data from the file, with extra stuff
                       // such as the preamble text and the separator
                       // lines removed.

    try {
      showStatus("Retrieving openings.txt...");
      fileUrl = new URL(getCodeBase(),"openings.txt");
      InputStream infile = fileUrl.openStream();

      BufferedReader ds = new BufferedReader(new InputStreamReader(infile));

      boolean recording = false;
      int     bar_count = 0;

      // read file, but filter out un-needed junk.
      // also put the delimiter "&&&" to separate
      // frequency section and sequence section

      for (;;) {
        String line = ds.readLine();

        if (line==null)
          break;

        boolean is_line = false;

        if (line.length()>=2) {
          String first_two = line.substring(0,2);

          is_line = first_two.equals("--") || first_two.equals("..");
        }

        if (recording && line.length()>0 && !is_line)
          data += line.trim() + "\n";  // record this line

        if (is_line) {
          if (++bar_count==2)  // we found two lines "--", start recording
            recording = true;
          else if (bar_count==3) // we found a third line, stop
            recording = false;
          else if (bar_count==4) { // found fourth, start again.
            recording = true;
            data += "&&&\n";       // append &&& to indicate sequence section
          }
        }
      }
    }
    catch (MalformedURLException mue) {
      System.out.println("Bad filename.");
      mue.printStackTrace();
    }
    catch (IOException ioe) {
      System.out.println("Data IO exception.");
      ioe.printStackTrace();
    }

    // create a list of nodes from data

    _alpha_openings = new OTNode();  // the _alpha_openings node will
                                     // contain no data, it is only
                                     // used as the top node
    {
        // put the data through a string tokenizer with the given delimiters.
      StringTokenizer st   = new StringTokenizer(data," \n\t\r*",true);
      boolean in_name      = true;
      boolean seq_section  = false;
      boolean f_count_done = false;
      String  last_token   = "";
      String  name         = "";
      String  moves        = "";
      OTNode  new_node     = null;

      while (st.hasMoreTokens()) {  // while there's more stuff
        String token = st.nextToken();

        if (token.equals("&&&"))
          seq_section = true;      // we've changed to the sequence section
        else {
          if (!seq_section) {      // still parsing frequency section
            if (in_name) {
              boolean name_ended = false;

              // the end of the name is always two spaces except for
              // Melnikov, Bat which ends in "(t3)"

              if (token.equals(" ") && last_token.equals(" "))
                name_ended = true;
              else if (token.equals("(t3)"))
                name_ended = true;

              if (!name_ended) {  // still parsing name
                if (!token.equals("*"))
                  name += token;  // don't add "*" to names
              }
              else {
                // name is complete... add a node for the opening
                in_name      = false;
                f_count_done = false;
                name = name.trim();

                Opening opening  = new Opening();
                        new_node = new OTNode();

                opening.setName(name);
                new_node.setData(opening);
              }
            }
            else { // parsing occurance statistics of frequency section
              if (token.equals("\n")) { // if done with line
                in_name = true;         // reinit state variables.
                name = "";
              }
              else if (!token.equals(" ") && !token.equals("\t")) {
                if (!f_count_done) { // haven't read freq count yet, parse it
                  new_node.data().setFrequency(Integer.parseInt(token));
                  f_count_done = true;
                }
                else { // parse freq percentage
                  String percent = token.substring(0,token.length()-1);
                  new_node.data().setFrequencyPercent(
                                    Double.valueOf(percent).doubleValue()
                                  );

                  // done parsing this opening's frequency, insert new_node
                  // into the _alpha_openings linked list sorted by
                  // frequency.  the highest frequency goes first.

                  if (_alpha_openings.next()==null)  // it's the first one
                    _alpha_openings.setNext(new_node);
                  else {
                    OTNode n = _alpha_openings;

                    while (n.next()!=null) {
                      OTNode n_next = (OTNode)n.next();

                      if (n_next.data().frequency()<new_node.data().frequency())
                        break;  // found the node with a lower frequency

                      n = n_next;
                    }

                    // insert the node

                    if (n.next()==null && 
                        n.data().frequency()>new_node.data().frequency()) {
                      n.setNext(new_node);
                    }
                    else {
                      new_node.setNext((OTNode)n.next());
                      n.setNext(new_node);
                    }
                  }
                }
              }
            }
          }
          else {  // parsing move sequence section
            in_name = false;
            name = "";

            if (moves.length()==0) {   // set the moves string
              if (!token.equals("\n"))
                moves = token;
            }
            else {  // already set the moves, must be the opening name then
              name = _parseName(st);

              // determine if this move/name has already been read.
              // there are multiple move sequences for each name for the
              // different orientations.  We are only storing the move
              // sequence for the first orientation.  We are deriving the
              // move sequences for the other orientations using the
              // "orientations.txt" file.

              OTNode x      = (OTNode)_alpha_openings.next();
              OTNode target = null;

              while (x!=null) {
                if (x.data().name().equals(name)) {  // found the node
                  if (x.data().moves().length()==0)  // if no move yet
                    target = x;                      // for the node

                  break;
                }

                x = (OTNode)x.next();
              }

              if (target!=null)  // store the move if it hasn't been yet
                target.data().setMoves(moves);

              moves = "";  // done with this move
            }
          }
        }

        last_token = token;
      }
    }

    // we're done parsing the data.  now we have to make the data into a
    // heirarchy that is structured based on continuations.  there are nodes
    // in the _openings tree for each individual move.  for example, the
    // "Diagonal" opening "C4c3" has a parent node with the move "C4", but
    // being that "C4" is not an opening, it has no name in its node.  The
    // "C4" move's parent is the _openings node.  The "X Square" opening,
    // "C4c3D3c5B2" would be placed using the following logic:  The first
    // move of the sequence, "C4" would be searched for in the tree.  It
    // would be found, then the second move "c3" would be search for from the
    // found node.  It would also be found (the "Diagonal" opening's node).
    // Then the next move, "D3", would be search for from the found node but
    // it would not be found so a new node with a move of "D3" is added as
    // a child to the previously found node.  This process is repeated until
    // a node for each move of the opening is in the tree.  The opening's
    // data is stored in its final move's node.

    {
      OTNode node = (OTNode)_alpha_openings.next();

      _openings = new OTNode();   // _openings will never contain data or
      _openings.setMove("XX");    // sibling nodes, it's just used as the
                                  // root node in the tree.

      while (node!=null) {
        // search for spot in the tree
        {
          String moves  = node.data().moves();
          int    m_i    = 0;
          OTNode parent = _openings;  // _openings is the root of the tree
          OTNode child  = (OTNode)parent.child();
          String move   = moves.substring(m_i,m_i+2); // the first move

          for (;;) {
            while (child!=null) {  // while there are children
              if (child.move().equals(move)) {  // child is for current move
                m_i += 2;

                if (m_i>=moves.length()) {  // if it's the last move
                  if (child.data()!=null)     // this shouldn't happen
                    System.out.println("Duplicate move "+child.data().moves());
                  else
                    child.setData(node.data());  // store the node's data

                  break;
                }

                parent = child;                      // traverse the tree
                child  = (OTNode)parent.child();
                move   = moves.substring(m_i,m_i+2); // get the next move
              }
              else  // child is not for the current move
                child = (OTNode)child.next(); // traverse to next child
            }

            if (m_i>=moves.length())  // didn't need to add any nodes for
              break;                  // for move, we're done with this one.

            // child is null, couldn't find move... add it

            OTNode new_node = new OTNode();

            new_node.setMove(move);
            parent.addChild(new_node);
            child = new_node;

            m_i += 2;

            if (m_i>=moves.length()) {  // placed the last move, store data
              if (child.data()!=null)     // this shouldn't happen
                System.out.println("Duplicate move "+child.data().moves());
              else
                child.setData(node.data());

              break;
            }

            parent = child;                     // traverse to new node
            child  = (OTNode)parent.child();
            move = moves.substring(m_i,m_i+2);  // get next move
          }
        }

        node = (OTNode)node.next();  // go to next node in _alpha_openings
      }
    }
  }


  /**
   *
   * Updates the push buttons' enabled/disabled states.
   *
   */
  protected void _updateBtns()
  {
    if (initialSetup()) {
      _first_btn.setEnabled(false);
      _next_btn.setEnabled(false);
      _prev_btn.setEnabled(false);
      _last_btn.setEnabled(false);
      _play_btn.setEnabled(false);
      _stop_btn.setEnabled(false);
      return;
    }

    _first_btn.setEnabled(true);
    _last_btn.setEnabled(true);
    _prev_btn.setEnabled(_played_moves.length()>0);
    _next_btn.setEnabled(_unplayed_moves.length()>0);

    if (_player_thread==null) {
      _stop_btn.setEnabled(false);
      _play_btn.setEnabled(true);
    }
    else {
      _stop_btn.setEnabled(true);
      _play_btn.setEnabled(false);
    }
  }


  /**
   *
   * Clears the board, initializes the unplayed and played moves.  The
   * played and unplayed moves are used so that the user can traverse
   * through an opening's individual moves.
   *
   */
  protected void _resetBoard()
  {
    _board.clear();

    if (_opening.data()==null)
      _unplayed_moves = "";
    else
      _unplayed_moves = _opening.data().moves();

    _played_moves = "";
  }


  /**
   *
   * Translates a row/col pair into a new row/pair for the currently
   * selected board orientation.
   *
   * @param row  the row to translate.  this parameter will be modified
   * @param col  the column to translate.  this parameter will be modified
   *
   */
  protected void _translateRowColumn(IntHolder row,IntHolder col)
  {
    int new_index = _curr_lookup[row.value*8+col.value]-1;
    row.value = new_index/8;
    col.value = new_index%8;
  }


  /**
   *
   * Translates a single move into a row/col pair for the currently selected
   * board orientation.
   *
   * @param move  the move to translate
   * @param row   the resulting row
   * @param col   the resulting column
   * @return true if the move would result in placing a white disc, false if
   * a black disc would be placed
   *
   */
  protected boolean _getMovePos(String move,IntHolder row,IntHolder col)
  {
    char c_char = move.charAt(0);
    char r_char = move.charAt(1);

    boolean white_move = c_char>='a';

    row.value = Character.digit(r_char,10)-1;

    if (row.value== -1)
      System.out.println("Invalid row: " + r_char);

    switch (c_char) {
      case 'a':
      case 'A':
        col.value = A;
        break;
      case 'b':
      case 'B':
        col.value = B;
        break;
      case 'c':
      case 'C':
        col.value = C;
        break;
      case 'd':
      case 'D':
        col.value = D;
        break;
      case 'e':
      case 'E':
        col.value = E;
        break;
      case 'f':
      case 'F':
        col.value = F;
        break;
      case 'g':
      case 'G':
        col.value = G;
        break;
      case 'h':
      case 'H':
        col.value = H;
        break;
      default:
        System.out.println("Invalid column: " + c_char);
    }

    if (_curr_lookup!=_o1_lookup)
      _translateRowColumn(row,col);

    return white_move;
  }


  /**
   *
   * Performs one move by translating the move into a row/col for the
   * currently selected board orientation.  The played moves is updated
   * to reflect the "playing" of this move.
   *
   * @param move  the move to make
   *
   */
  protected void _makeMove(String move)
  {
    boolean white_move;
    int     r,c;

    {
      IntHolder  r_h  = new IntHolder();
      IntHolder  c_h  = new IntHolder();

      white_move = _getMovePos(move,r_h,c_h);
      r = r_h.value;
      c = c_h.value;
    }

    _board.placeDisc(r,c,white_move);
    _played_moves = _played_moves.concat(move);
  }


  /**
   *
   * Makes a number of moves and updates _unplayed_moves.
   *
   * @param num_to_make  the number of moves to make
   *
   */
  protected void _makeMoves(int num_to_make)
  {
    while (num_to_make>0) {
      _makeMove(_unplayed_moves.substring(0,2));
      _unplayed_moves = _unplayed_moves.substring(2);
      --num_to_make;
    }
  }


  /**
   *
   * Replays the current move of the selected opening.  This is only used
   * when the board orientation changes.
   *
   */
  protected void _currMove()
  {
    int n_to_play = _played_moves.length()/2;
    _resetBoard();
    _makeMoves(n_to_play);
  }


  /**
   *
   * Resets the board and makes one fewer move of the selected opening than
   * previously was shown.
   *
   */
  protected void _prevMove()
  {
    int n_to_play = _played_moves.length()/2-1;
    _resetBoard();
    _makeMoves(n_to_play);
  }


  /**
   *
   * Makes one additional move of the selected opening.
   *
   */
  protected void _nextMove()
  {
    _makeMoves(1);
  }


  /**
   *
   * Makes the remaining number of moves in the selected opening.
   *
   */
  protected void _lastMove()
  {
    _makeMoves(_unplayed_moves.length()/2);
  }


  /**
   *
   * Returns the tree node for the given move sequence.
   *
   * @param moves  the move sequence to find
   * @return the tree node for the move sequence
   *
   */
  protected OTNode _findTreeNode(String moves)
  {
    OTNode node = (OTNode)_openings.child();
    int    m_i  = 0;
    String move = moves.substring(m_i,m_i+2);

    for (;;) {
      if (node.move().equals(move)) {
        m_i += 2;

        if (m_i>=moves.length())
          break;

        move = moves.substring(m_i,m_i+2);
        node = (OTNode)node.child();
      }
      else
        node = (OTNode)node.next();

      if (m_i>=moves.length())
        break;
    }

    return node;
  }


  /**
   *
   * Fills the list with all of the openings, sorted by frequency.
   *
   */
  protected void _fillAlphaList()
  {
    _list_panel.removeAll();

    OTNode n = (OTNode)_alpha_openings.next();

    while (n!=null) {
      _list_panel.addItem(n);
      n = (OTNode)n.next();
    }
  }


  /**
   *
   * Fills the list with the continuations of the currently selected opening.
   *
   */
  protected void _fillContinuationsList()
  {
    _list_panel.removeAll();

    OTNode node = null;

    if (_opening.data()==null)
      node = _openings;
    else
      node = _findTreeNode(_opening.data().moves());

    _list_panel.addContinuationItems((OTNode)node);
    _opening = node;
  }


  /**
   *
   * Returns the parent opening of the given opening.
   *
   * @param opening the node of the opening to query
   * @return the node of the opening that is the parent of the given
   * opening's node
   *
   */
  protected OTNode _prevOpening(OTNode opening)
  {
    if (opening==null)
      return _openings;

    OTNode n = (OTNode)opening.parent();

    while (n!=_openings) { // while not at root of the tree
      if (n.data()!=null)  // found an opening's node, stop looking
	    break;

      n = (OTNode)n.parent();  // traverse to the parent
    }

    return n;
  }


  /**
   *
   * Translates a move sequence to the current board orientation.
   *
   * @param moves  move sequence for the base orientation
   * @return       a new move sequence for the current board orientation
   *
   */
  protected String _translateMovesString(String moves)
  {
    if (_curr_lookup==_o1_lookup)  // its in the base orientation,
      return moves;                // no need to translate

    String new_moves = "";

    int start = 0;  // index into the moves string

    for (;;) {
      if (start>=moves.length())
        break;

      char cch = moves.charAt(start);

      boolean is_upper = Character.isUpperCase(cch);

      int col;

      if (is_upper) {  // figure out the integer value for the column
        col = cch-'A';
      }
      else {
        col = cch-'a';
      }

      // extract the value for the row

      int row;

      {
        char rch = moves.charAt(start+1);

        row = Character.digit(rch,10)-1;
      }

      int new_r;
      int new_c;

      // figure out the orientation's cooresponding row/col

      {
        int new_index = _curr_lookup[row*8+col]-1;

        new_r = new_index/8;
        new_c = new_index%8;
      }

      // translate the new column back to a character, case is important

      if (is_upper) {
        cch = (char)('A'+new_c);
      }
      else {
        cch = (char)('a'+new_c);
      }

      // append the resulting characters to the string

      new_moves += cch;
      new_moves += (new_r+1);
      start += 2;
    }

    return new_moves;
  }


  /**
   *
   * Updates the on-screen label with the move string of the selected
   * opening that has been translated using the current board orientation.
   *
   */
  protected void _updateMoveSequenceLabel()
  {
    _moves_label.setText(_translateMovesString(_opening.data().moves()));
  }


  // PUBLIC


  /**
   *
   * Button listener for the play,stop,first,prev,next and last buttons.
   *
   * Note that javadoc 1.1.x does not generate docs for inner classes.
   *
   * @version 1.00
   * @author Bill Lorton
   *
   */
  public class HandleButton implements ActionListener {
    private Othello _applet;

    HandleButton() { }
    HandleButton(Othello a) { _applet = a; }

    public void actionPerformed(ActionEvent e)
    {
      if (e.getSource()==_play_btn) {
	_player_thread = new PlayerThread(_applet);
	_player_thread.start();
	_updateBtns();
      }
      else if (e.getSource()==_stop_btn) {
	_player_thread.stop();
	_player_thread = null;
	_updateBtns();
      }
      else if (e.getSource()==_first_btn) {
	_resetBoard();
	_board.repaint();
	_updateBtns();
      }
      else if (e.getSource()==_prev_btn) {
	_prevMove();
	_board.repaint();
	_updateBtns();
      }
      else if (e.getSource()==_next_btn) {
	_nextMove();
	_board.repaint();
	_updateBtns();
      }
      else if (e.getSource()==_last_btn) {
	_lastMove();
	_board.repaint();
	_updateBtns();
      }
    }
  }

  /**
   *
   * Checkbox listener for the orientation check boxes.
   *
   * Note that javadoc 1.1.x does not generate docs for inner classes.
   *
   * @version 1.00
   * @author Bill Lorton
   *
   */
  public class HandleCheckbox implements ItemListener {
    private Checkbox _check_box;

    public HandleCheckbox(Checkbox c) { _check_box = c; }

    public void itemStateChanged(ItemEvent e)
    {
      if (e.getStateChange()==e.DESELECTED)
	return; // don't care about deselections

      if (_check_box==_o1_box) {
	_curr_lookup = _o1_lookup;

	if (!initialSetup()) {
	  _currMove();
	  _updateMoveSequenceLabel();
	  _board.repaint();
	}
      }
      else if (_check_box==_o2_box) {
	_curr_lookup = _o2_lookup;

	if (!initialSetup()) {
	  _currMove();
	  _moves_label.setText(_translateMovesString(_opening.data().moves()));
	  _board.repaint();
	}
      }
      else if (_check_box==_o3_box) {
	_curr_lookup = _o3_lookup;

	if (!initialSetup()) {
	  _currMove();
	  _moves_label.setText(_translateMovesString(_opening.data().moves()));
	  _board.repaint();
	}
      }
      else if (_check_box==_o4_box) {
	_curr_lookup = _o4_lookup;

	if (!initialSetup()) {
	  _currMove();
	  _updateMoveSequenceLabel();
	  _board.repaint();
	}
      }
    }
  }

  /**
   *
   * Creates the user interface, reads "openings.txt" and "orientations.txt",
   * fills the openings list.
   *
   */
  public void init()
  {
    Font label_font = new Font("TimesRoman",Font.BOLD,18);
    Font value_font = new Font("Corier",Font.PLAIN,14);

    setBackground(Color.white);

    setLayout(new GridBagLayout());

    GBHelper gbh = new GBHelper();

    _board = new Board(this);
    _board.init();

    gbh.constrain(this,_board,0,0,1,1);

    _right_panel = new Panel();
    _right_panel.setLayout(new GridBagLayout());

    _list_panel = new ListPanel(this);

    gbh.constrainFill(_right_panel,_list_panel,0,0,1,2);

    Panel orient_panel = new Panel();
    orient_panel.setLayout(new GridBagLayout());

    Label orient_label = new Label("Board Orientation:");
    gbh.constrainFill(orient_panel,orient_label,0,0,1,1);

    Panel radio_panel = new Panel();

    radio_panel.setLayout(new FlowLayout(FlowLayout.LEFT,0,0));

    _orientation_group = new CheckboxGroup();

    _o1_box = new Checkbox("1",_orientation_group,true);
    _o1_box.addItemListener(new HandleCheckbox(_o1_box));
    radio_panel.add(_o1_box);
    _o2_box = new Checkbox("2",_orientation_group,false);
    _o2_box.addItemListener(new HandleCheckbox(_o2_box));
    radio_panel.add(_o2_box);
    _o3_box = new Checkbox("3",_orientation_group,false);
    _o3_box.addItemListener(new HandleCheckbox(_o3_box));
    radio_panel.add(_o3_box);
    _o4_box = new Checkbox("4",_orientation_group,false);
    _o4_box.addItemListener(new HandleCheckbox(_o4_box));
    radio_panel.add(_o4_box);

    gbh.constrainFill(orient_panel,radio_panel,0,1,1,1);

    gbh.constrain(_right_panel,orient_panel,0,2,1,1,10,0,0,0);

    Panel btn_panel = new Panel();
    btn_panel.setLayout(new GridBagLayout());

    _play_btn = new Button("Play");
    _play_btn.addActionListener(new HandleButton(this));
    gbh.constrainFill(btn_panel,_play_btn,0,0,2,1);

    _stop_btn = new Button("Stop");
    _stop_btn.addActionListener(new HandleButton());
    gbh.constrainFill(btn_panel,_stop_btn,2,0,2,1);

    _first_btn = new Button("First");
    _first_btn.addActionListener(new HandleButton());
    gbh.constrain(btn_panel,_first_btn,0,1,1,1,5,0,0,0);

    _prev_btn  = new Button("Prev");
    _prev_btn.addActionListener(new HandleButton());
    gbh.constrain(btn_panel,_prev_btn,1,1,1,1,5,0,0,0);

    _next_btn  = new Button("Next");
    _next_btn.addActionListener(new HandleButton());
    gbh.constrain(btn_panel,_next_btn,2,1,1,1,5,0,0,0);

    _last_btn  = new Button("Last");
    _last_btn.addActionListener(new HandleButton());
    gbh.constrain(btn_panel,_last_btn,3,1,1,1,5,0,0,0);

    gbh.constrain(_right_panel,btn_panel,0,3,1,1,0,0,0,0);

    gbh.constrain(this,_right_panel,1,0,1,1,10,10,0,0);

    Panel bottom_panel = new Panel();
    bottom_panel.setLayout(new GridBagLayout());

    Panel info_panel = new Panel();
    info_panel.setLayout(new GridBagLayout());

    Label opening_label = new Label("Current Opening:");

    opening_label.setFont(label_font);
    gbh.constrain(info_panel,opening_label,0,0,1,1);

    _name_label = new Label("Initial Setup                                           ");
    _name_label.setFont(value_font);
    gbh.constrain(info_panel,_name_label,0,1,4,1,0,10,0,0);

    Label moves_label = new Label("Move Sequence:");

    moves_label.setFont(label_font);
    gbh.constrain(info_panel,moves_label,0,2,1,1);

    _moves_label =
      new Label("                                                                           ");
    _moves_label.setFont(value_font);

    gbh.constrain(info_panel,_moves_label,0,3,4,1,0,10,0,0);

    Label frequency_label = new Label("Frequency:");
    frequency_label.setFont(label_font);
    gbh.constrain(info_panel,frequency_label,0,4,1,1);

    _frequency_label = new Label("100%          ");
    _frequency_label.setFont(value_font);
    gbh.constrain(info_panel,_frequency_label,0,5,4,1,0,10,0,0);

    gbh.constrain(bottom_panel,info_panel,0,0,1,1,10,40,0,0);

    gbh.constrain(this,bottom_panel,0,1,1,1);

    validate();

    _readOrientations();
    _readOpenings();
    _opening = _alpha_openings;

    _fillAlphaList();
    _resetBoard();
    _updateBtns();

    _player_thread = null;
  }


  /**
   *
   * Returns whether or not the current state of the board is the initial
   * setup that contains only the four inner discs.
   *
   * @return true if the current opening is the initial setup
   *
   */
  public boolean initialSetup()
  {
    return _opening.data()==null || _opening.data().moves().length()==0;
  }


  /**
   *
   * Called when an opening is selected from the list, this method updates
   * the interface to display the new opening.  If the continuations only
   * toggle is set, the list is also updated to just show the child
   * openings of the newly selected opening.
   *
   * @param n                   the node of the opening that was selected
   * @param continuations_only  the current state of the continuations only
   * toggle
   */
  public void listSelectCallback(OTNode n,boolean continuations_only)
  {
    if (_player_thread!=null) {  // if playing, stop
      _player_thread.stop();
      _player_thread = null;
    }

    _opening = n;

    if (continuations_only)
      _fillContinuationsList();

    _name_label.setText(_opening.data().name());
    _moves_label.setText(_translateMovesString(_opening.data().moves()));
    _frequency_label.setText(
                       Double.toString(_opening.data().frequencyPercent())+"%"
                     );
    _resetBoard();
    _lastMove();
    _board.repaint();
    _updateBtns();
  }


  /**
   *
   * Updates the list when the user clicks on the Continuations Only
   * toggle button.  If the previously selected item in the list is in
   * the new list, that item selected will be selected in the new list.
   *
   * @param continuations_only  the new value of the continuations toggle
   */
  public void continuationsOnlyChangedCallback(boolean continuations_only)
  {
    if (continuations_only)
      _fillContinuationsList();
    else {
      _fillAlphaList();

      // if a tree node was selected, find the associated alpha node

      if (_opening!=null)
        _list_panel.selectItem(_opening);
    }
  }


  /**
   *
   * Finds the parent opening for the current opening and then displays
   * a list of all child openings of that parent opening.  Updates the
   * display to indicate the newly selected opening.  If the previously
   * selected opening was a top level opening then display "Initial
   * Setup" as the new selection.  This method can only be called in
   * "Continuations Only" mode.
   *
   * @return  non-zero if the new opening exists, zero if the it is
   * the initial setup
   */
  public boolean previousOpening()
  {
    if (_player_thread!=null) {
      _player_thread.stop();
      _player_thread = null;
    }

    OTNode n = _prevOpening(_opening);

    _opening = n;
    _fillContinuationsList();

    if (_opening.data()==null) {
      _name_label.setText("Initial Setup");
      _moves_label.setText(" ");
      _frequency_label.setText("100%");
    }
    else {
      _name_label.setText(_opening.data().name());
      _moves_label.setText(_opening.data().moves());
      _frequency_label.setText(
                         Double.toString(_opening.data().frequencyPercent())+"%"
		       );
    }

    _resetBoard();
    _lastMove();
    _board.repaint();
    _updateBtns();
    return _opening.data()!=null;
  }


  /**
   *
   * Returns whether or not the current move is the last of the selected
   * opening.
   *
   * @return non-zero if the current move is the last of the selected
   * opening
   */
  public boolean atLastMove()
  {
    return _unplayed_moves.length()==0;
  }


  /**
   *
   * Sets the current move of the selected opening to the first move.
   *
   */
  public void firstMove()
  {
    _resetBoard();
    _board.repaint();
    _updateBtns();
  }


  /**
   *
   * Sets the current move of ths selected opening to the next move.
   *
   */
  public void nextMove()
  {
    _nextMove();
    _board.repaint();
    _updateBtns();
  }

  /**
   *
   * Stops playing through the moves of the selected opening.
   *
   */
  public void stop()
  {
    if (_player_thread!=null) {
      _player_thread.stop();
      _player_thread = null;
      _updateBtns();
    }
  }
}



